@@ -1,5 +1,8 @@
 from .users import User
-from .session import Session
+from .session import Session, GuestSession
 from .works import Work
 from .search import Search
+from .comments import Comment
 from . import utils
+
+VERSION = "1.2.1"
\ No newline at end of file
