@@ -4,6 +4,8 @@ from functools import cached_property
 import requests
 from bs4 import BeautifulSoup
 
+from . import utils
+
 
 class User:
     """
@@ -21,6 +23,19 @@ class User:
         self.soup_works = self.request("https://archiveofourown.org/users/%s/works?page=1"%(username))
         self.loaded_page = 1
         self.soup_profile = self.request("https://archiveofourown.org/users/%s/profile"%username)
+        
+    def get_avatar(self):
+        """Returns a tuple containing the name of the file and its data
+
+        Returns:
+            tuple: (name: str, img: bytes)
+        """
+        
+        icon = self.soup_profile.find("p", {"class": "icon"})
+        src = icon.img.attrs["src"]
+        name = src.split("/")[-1].split("?")[0]
+        img = requests.get(src).content
+        return name, img
 
     @cached_property
     def works(self):
@@ -30,8 +45,10 @@ class User:
             int: Number of works
         """
 
-        n = self.soup_works.find("h2", {'class': 'heading'}).string.strip().split(" ")[0]
-        return int(n)
+        div = self.soup_works.find("div", {'id': 'inner'})
+        span = div.find("span", {'class': 'current'}).getText().replace("(", "").replace(")", "")
+        n = span.split(" ")[1]
+        return int(self.str_format(n))   
 
     @cached_property
     def npages(self):
@@ -98,8 +115,10 @@ class User:
         """
 
         req = requests.get(url)
+        if req.status_code == 429:
+            raise utils.HTTPError("We are being rate-limited. Try again in a while or reduce the number of requests")
         content = req.content
-        soup = BeautifulSoup(content, "html.parser")
+        soup = BeautifulSoup(content, "lxml")
         return soup
 
     @staticmethod
