@@ -5,6 +5,7 @@ import requests
 from bs4 import BeautifulSoup
 
 from . import utils
+from .comments import Comment
 
 
 class Work:
@@ -12,23 +13,24 @@ class Work:
     AO3 work object
     """
 
-    def __init__(self, workid):
+    def __init__(self, workid, session=None):
         """Creates a new AO3 work object
 
         Args:
             workid (int): AO3 work ID
+            session (AO3.Session, optional): Used to access restricted works
 
         Raises:
-            requests.HTTPError: Raised if the work wasn't found
+            utils.InvalidIdError: Raised if the work wasn't found
         """
 
+        self._session = session
         self.chapter_ids = []
         self.chapter_names = []
         self.workid = workid
         self.soup = self.request("https://archiveofourown.org/works/%i?view_adult=true"%workid)
         if "Error 404" in self.soup.text:
-            raise requests.HTTPError("Error 404: cannot find work")
-        
+            raise utils.InvalidIdError("Cannot find work")
 
     def get_chapter_text(self, chapter):
         """Gets the chapter text from the specified chapter.
@@ -58,22 +60,155 @@ class Work:
         else:
             raise utils.UnloadedError("Work.load_chapters() must be called first")
     
-    def load_chapters(self):
+    def load_chapters(self, session=None):
         """
         Loads the urls for all chapters
         """
         
-        if self.chapters > 1:
+        if not self.oneshot:
             navigate = self.request("https://archiveofourown.org/works/%i/navigate?view_adult=true"%self.workid)
             all_chapters = navigate.find("ol", {'class': 'chapter index group'})
+            if all_chapters is None:
+                raise utils.AuthError("This work is only available to registered users of the Archive")
             self.chapter_ids = []
             self.chapter_names = []
-            for chapter in all_chapters.find_all("li"):
+            for chapter in all_chapters.findAll("li"):
                 self.chapter_ids.append(chapter.a['href'].split("/")[-1])
                 self.chapter_names.append(chapter.a.string)
         else:
             self.chapter_ids = [""]
             self.chapter_names = [self.title]
+            
+    def download(self, filetype="PDF"):
+        """Downloads this work
+
+        Args:
+            filetype (str, optional): Desired filetype. Defaults to "PDF".
+
+        Raises:
+            utils.DownloadError: Raised if there was an error with the download
+            utils.UnexpectedResponseError: Raised if the filetype is not available for download
+
+        Returns:
+            bytes: File content
+        """
+        
+        download_btn = self.soup.find("li", {"class": "download"})
+        for download_type in download_btn.findAll("li"):
+            if download_type.a.getText() == filetype:
+                url = f"https://archiveofourown.org/{download_type.a.attrs['href']}"
+                req = requests.get(url)
+                if req.status_code == 429:
+                    raise utils.HTTPError("We are being rate-limited. Try again in a while or reduce the number of requests")
+                if not req.ok:
+                    raise utils.DownloadError("An error occurred while downloading the work")
+                return req.content
+        raise utils.UnexpectedResponseError(f"Filetype '{filetype}' is not available for download")
+    
+    def get_comments(self, chapter=None, maximum=None):
+        """Returns a list of all threads of comments in the specified chapter. This operation can take a very long time.
+        Because of that, it is recomended that you set a maximum number of comments. 
+        Duration: ~ (0.13 * n_comments) seconds or 2.9 seconds per comment page
+
+        Args:
+            chapter (int/str, optional): Chapter number, only required if work is not a oneshot. Defaults to None.
+            maximum (int, optional): Maximum number of comments to be returned. None -> No maximum
+
+        Raises:
+            ValueError: Invalid chapter number
+            IndexError: Invalid chapter number
+            utils.UnloadedError: Work.load_chapters() must be called first
+
+        Returns:
+            list: List of comments
+        """
+        
+        if self.oneshot:
+            chapter_id = self.workid
+        else:
+            if chapter is None:
+                raise IndexError("chapter cannot be 'None'")
+            if chapter <= 0 or chapter >= self.chapters:
+                raise IndexError("Invalid chapter number")
+            if len(self.chapter_ids) != self.chapters:
+                raise utils.UnloadedError("Work.load_chapters() must be called first")
+            
+            chapter_id = self.chapter_ids[chapter-1]
+            
+        url = f"https://archiveofourown.org/comments/show_comments?page=%d&chapter_id={chapter_id}"
+        soup = self.request(url%1)
+        
+        pages = 0
+        div = soup.find("div", {"id": "comments_placeholder"})
+        ol = div.find("ol", {"class": "pagination actions"})
+        if ol is None:
+            pages = 1
+        else:
+            for li in ol.findAll("li"):
+                if li.getText().isdigit():
+                    pages = int(li.getText())   
+        
+        comments = []
+        for page in range(pages):
+            if page != 0:
+                soup = self.request(url%(page+1))
+            ol = soup.find("ol", {"class": "thread"})
+            for li in ol.findAll("li", {"role": "article"}, recursive=False):
+                if maximum is not None and len(comments) >= maximum:
+                    return comments
+                id_ = int(li.attrs["id"][8:])
+                comments.append(Comment(id_, chapter_id))
+        return comments
+    
+    def leave_kudos(self, session):
+        """Leave a 'kudos' in this work
+
+        Args:
+            session (AO3.Session/AO3.GuestSession): session object
+
+        Raises:
+            utils.UnexpectedResponseError: Unexpected response received
+            utils.InvalidIdError: Invalid workid (work doesn't exist)
+            utils.AuthError: Invalid authenticity token
+
+        Returns:
+            bool: True if successful, False if you already left kudos there
+        """
+        
+        return utils.kudos(self.workid, session)
+    
+    def comment(self, chapter, comment_text, session, email="", name=""):
+        """Leaves a comment on this work
+
+        Args:
+            chapter (int): Chapter number
+            comment_text (str): Comment text
+            session (AO3.Session, optional): Session object. Defaults to None (posts anonimously).
+
+        Raises:
+            IndexError: Invalid chapter number
+            utils.UnloadedError: Couldn't load chapter ids. Call Work.load_chapters() first
+
+        Returns:
+            requests.models.Response: Response object
+        """
+        
+        if chapter < 1 or chapter > self.chapters:
+            raise IndexError(f"Invalid chapter number")
+        
+        if len(self.chapter_ids) != self.chapters:
+            raise utils.UnloadedError("Work.load_chapters() must be called first")
+        
+        if self.chapters == 1:
+            chapterid = self.workid
+        else:
+            chapterid = self.chapter_ids[chapter-1]
+            
+        return utils.comment(chapterid, comment_text, self.oneshot, email=email, name=name)
+    
+    @property
+    def oneshot(self):
+        return self.chapters == 1
 
     @cached_property
     def authors(self):
@@ -351,8 +486,7 @@ class Work:
 
         return "https://archiveofourown.org/works/%i"%self.workid      
 
-    @staticmethod
-    def request(url):
+    def request(self, url):
         """Request a web page and return a BeautifulSoup object.
 
         Args:
@@ -363,9 +497,14 @@ class Work:
             bs4.BeautifulSoup: BeautifulSoup object representing the requested page's html
         """
 
-        req = requests.get(url)
+        if self._session is None:
+            req = requests.get(url)
+        else:
+            req = self._session.session.get(url)
+        if req.status_code == 429:
+            raise utils.HTTPError("We are being rate-limited. Try again in a while or reduce the number of requests")
         content = req.content
-        soup = BeautifulSoup(content, "html.parser")
+        soup = BeautifulSoup(content, "lxml")
         return soup
 
     @staticmethod
